# tcma.py
#
# Copyright (C): 2008 Per B. Sederberg
# Authors: Per B. Sederberg
# URL: http://memory.psych.upenn.edu/publications
#
# Distributed under the terms of the GNU General Public License
# (GPL). See the license.txt that came with this file.
#

from __future__ import division
import sys
from numpy import *
import random as rnd

import pdb

# local imports
# see if they have compiled the helper
try:
    import helper
    runHelper = True
except:
    runHelper = False
    print "Compiled accumulator code not loaded. Using slower version..."

def ismember(a,b):
    ainb = zeros(len(a),dtype=bool)
    for item in b:
        ainb = ainb | (a==item)
    return ainb

class TCMA:
    def __init__(self,params):
	"""
	"""
	self.numUnits = getValFromDict(params,'numUnits',200)

	# set up tcm params
	self.B_enc = getValFromDict(params,'B_enc',.65) # drift
	self.B_ret = getValFromDict(params,'B_ret',self.B_enc) # drift
	self.B_dist = getValFromDict(params,'B_dist',.65) # drift
	self.Gft = getValFromDict(params,'Gft',.5)
	self.Gft_ret = getValFromDict(params,'Gft_ret',.5)
	self.Gft_enc = getValFromDict(params,'Gft_enc',-1)
        if self.Gft_enc < 0:
            # set it to ret
            self.Gft_enc = self.Gft_ret
	self.Gtf = getValFromDict(params,'Gtf',.5)
        
        # primacy params
	self.pScale = getValFromDict(params,'pScale',0.)
	self.pDecay = getValFromDict(params,'pDecay',0.)

        # learning rate
	self.lRate = getValFromDict(params,'lRate',1.0)

        # accumulator params
	self.dt = getValFromDict(params,'dt',100.)
	self.tau = getValFromDict(params,'tau',1000.)
        # divide by sqrt(dt/tau) to match equation in paper
	self.eta = getValFromDict(params,'eta',.335) / sqrt(self.dt/self.tau)
	self.L = getValFromDict(params,'L',.02)
	self.K = getValFromDict(params,'K',.45)
	self.thresh = getValFromDict(params,'thresh',1.)
	self.minI = getValFromDict(params,'minI',0.)

        # recall duration
	self.recDuration = getValFromDict(params,'recDuration',45000)

        self.outputEncoding = getValFromDict(params,'outputEncoding',False)
        
	# call reset to set up matrixes
	self.Reset()
    
    def Reset(self):
	"""
	Reset the matrixes and the model in general.
	"""
	# set up the matrixes
	self.Mtf_fast = zeros([self.numUnits,self.numUnits],'f')
	self.Mft_fast = zeros([self.numUnits,self.numUnits],'f')
	self.Mtf_slow = identity(self.numUnits,'f')
	self.Mft_slow = identity(self.numUnits,'f')
	self.t = zeros(self.numUnits,'f')
	self.x = None
        self.learnRate = self.lRate
        self.last_iNoise = 1
	self.items = identity(self.numUnits,'f')
	self.curItem = 0
	self.listInd = []
        self.distInd = []
	self.recalled = []
	self.recTime = []
	self.numRepeats = 0

	# run for one distractor to initialize t
	self.Present(isDistract = True)

    def Present(self, itemInd = None, updateWeights = True, appendItem = True,
                isDistract = False, isEncoding = True, customB = None, appendDistract = False):
	"""
	"""
	
	# see if isDistract
	if isDistract:
	    # don't append item or update weights
	    appendItem = False
	    updateWeights = False
	    B = self.B_dist
	elif isEncoding:
	    B = self.B_enc
            Gft = self.Gft_enc
	else:
	    B = self.B_ret
            Gft = self.Gft_ret

        # if custom provided, use that Beta
	if not customB is None:
	    B = customB

	# get the item to present
	if itemInd is None:
	    # get the f from current item
	    f = self.items[self.curItem]
	    
	    # see if append item
	    if appendItem:
		self.listInd.append(self.curItem)

            if appendDistract:
                self.distInd.append(self.curItem)
                
	    # update to the next item
	    self.curItem = self.curItem+1
	else:
	    # get the item from the value passed in
	    f = self.items[itemInd]
	    
	    # see if append item
	    if appendItem:
		self.listInd.append(itemInd)

            if appendDistract:
                self.distInd.append(itemInd)
	
	# see if vary tIN strength of distractor
	if isDistract:            
	    # only use the slow
	    tIN = inner(self.Mft_slow,f)
	else:
	    # determine the tIN from both f to t matrixes
	    tIN_fast = inner(self.Mft_fast,f)

            # add in the slow component
	    tIN_slow = inner(self.Mft_slow,f)

	    tIN = (Gft*tIN_fast + (1-Gft)*tIN_slow)
            # make sure unit length
            #tIN = tIN/sqrt(inner(tIN,tIN))

	# make sure unit length
	tIN = tIN/sqrt(inner(tIN,tIN))

	# update t
	rho = sqrt(1+(B**2)*((dot(self.t,tIN)**2)-1)) - B*dot(self.t,tIN)
	self.t_last = self.t
	self.t = rho*self.t + B*tIN
	
	# see if update weights
	if updateWeights:
	    # update fast M, before updating t
	    # determine the primacy factor
            if isEncoding:
                primefact = self.pScale*exp(-self.pDecay*(len(self.listInd)-1)) + 1
            else:
                primefact = 1.0

            # do the weight update
            Dft_fast = outer(self.t_last,f)*self.learnRate
	    self.Mft_fast = self.Mft_fast + Dft_fast

            Dtf_fast = outer(f,self.t_last)*self.learnRate*primefact
	    self.Mtf_fast = self.Mtf_fast + Dtf_fast
	
    def Train(self,numItems,contDistract=False,listDef=None):
	"""
	"""
	# do some checks to make sure there are enough units
        
        # handle the listDef
        possibleItems = arange(self.curItem,self.curItem+numItems)
        if listDef is None:
            items = possibleItems[range(numItems)]
        else:
            # go by list definition
            if not len(listDef)==numItems:
                raise "Length of listDef must equal numItems."
            items = possibleItems[listDef]

        # set the current item
        self.curItem = max(items+1)

        # see if run continuous distract
        if contDistract:
            # present a distractor item
            self.Present(isDistract = True,appendDistract=True)

	# loop and train
	for i in items:
            # Present an item normally
            self.Present(itemInd=i)

	    # see if run continuous distract
	    if contDistract:
		# present a distractor item
                self.Present(isDistract = True,appendDistract=True)

	if not contDistract:
	    # present a distractor item
            self.Present(isDistract = True,appendDistract=True)

    def accumulate(self,x,I,recItems,randn_vals,tempEta,
                   dt_tau,sq_dt_tau):
        """
        """

        # get the num items
        numItems = len(x)

        # get the notInd
        notInd = recItems==0

        # get maxTime
        maxTime = randn_vals.shape[0]

        # loop over time
        recalledItem = False
        for t in xrange(maxTime):
            # get the lateral inhibition
            lx = x.sum()*ones(numItems) - x

            # get the randn range
            #ind = range(t*numItems,t*numItems+numItems)

            # calc new values
            x += (I - self.K*x - self.L*lx)*dt_tau + (tempEta*randn_vals[t,:]*sq_dt_tau)

            # make sure positive
            x[x<0] = 0.0

            # see if crossed thresh
            if any(x[notInd]>=self.thresh):
                # we've recalled one, so done
                recalledItem = True
                break

            # if already recalled and crossed, then reset to zero
            x[(x>=self.thresh) & ~notInd] = 0

        if not recalledItem:
            # must increment t b/c to make run like C++ for loop
            t+=1
            
        return t,x


    def RecallItem(self,maxTime):
	"""
	"""
	# have not recalled anything yet
	recInd = None
        
        # get unique list of item serial positions
        uniqueSP,uniqueItems = unique1d(self.listInd,True)

	# get not recalled ind
	notInd = nonzero(~ismember(uniqueSP,self.recalled))[0]
	if len(notInd) == 0:
	    # no more to recall
	    return recInd,None

	# get input to accumulators from the t to f matrix
	# combination of both fast and slow matrixes
	f_fast = inner(self.Mtf_fast,self.t)
	f_slow = inner(self.Mtf_slow,self.t)	
	I = f_fast*self.Gtf + (1-self.Gtf)*f_slow

	# only accumulate items from the list
        I = I[uniqueItems]

        if self.minI > 0:
            # ignore small inputs
            I[I<self.minI] = 0

            # measure of input noise	
            noiseSTD = std(I[I>=self.minI])
            noiseMEAN = mean(I[I>=self.minI]) 
            if noiseSTD == 0 or noiseMEAN == 0:
                sys.stderr.write('.')
                #iNoise = 1.
                iNoise = self.last_iNoise
            else:
                iNoise = noiseSTD/noiseMEAN
                self.last_iNoise = iNoise
        else:
            iNoise = 1.
                
	# Scale the input based on the CV
	I = I*(iNoise)
	tempEta = self.eta #*etaNoise

        # save the I and CV terms
        self.input.append(I)
        self.CV.append(iNoise)

	if self.x is None or len(self.x) != len(I):
	    # must allocate
	    self.x = zeros(len(I),'f')

        # Do some precalculations to speed up accumulators 
	numItems = len(uniqueItems)
	dt_tau = self.dt/self.tau
	sq_dt_tau = sqrt(dt_tau)
        randn_vals = random.randn(maxTime,numItems)
        
	# run the accumulator
        if runHelper:
            # set up matrix for external accumulator
            amat = empty([maxTime+3,numItems],float64)
            amat[0,:] = self.x  # accumulators
            amat[1,:] = I  # input to accum
            amat[2,:] = ones(numItems)  # recalled items
            amat[2,notInd] = 0
            amat[3:,:] = randn_vals  # randn vals to use

            i = helper.accumulate(amat,self.K,self.L,tempEta,self.thresh,
                                  dt_tau,sq_dt_tau)
        else:
            recItems = ones(numItems)
            recItems[notInd] = 0
            i,self.x = self.accumulate(self.x,I,recItems,randn_vals,tempEta,
                                       dt_tau,sq_dt_tau)

	if i<maxTime:
	    # we recalled something
	    # find out which one
            if runHelper:
                self.x = squeeze(amat[0,:])
	    nInd = nonzero(self.x[notInd] == max(self.x[notInd]))[0]

	    # make sure not get more than one
	    if len(nInd) > 1:
		nInd = nInd[argsort(rand(len(nInd)))[0]]
		
	    # get the itemno
	    itemNo = notInd[nInd]

	    # set the recalled x to zero
	    self.x[itemNo] = 0

	    # index into the list
	    recInd = squeeze(uniqueItems[itemNo])
	    
	    # call present on the recalled item to update context
            self.Present(itemInd=recInd,
                         updateWeights = self.outputEncoding, 
                         appendItem = False, isEncoding = False)

	    # add it for now
	    if itemNo < numItems:
		# is actual item
		self.recalled.append(squeeze(uniqueSP[itemNo]))
		# add the rectime
		self.recTime.append(self.numIterations+i)
		# reset the iterations
		self.numIterations = 0
	    else:
		# is intrusion
		self.numIterations = self.numIterations + i
	    
	# return what recalled, if anything, and the time elapsed
	return recInd,i

    def Test(self):
	"""
	"""
                
	# determine the total timesteps
	totTime = self.recDuration / self.dt
	timeRemaining = totTime

        self.curTime = 0
	self.numIterations = 0

	# recall items until time is up or no more to recall
	while timeRemaining > 0:
	    # try and recall an item
	    recItem,recTime = self.RecallItem(timeRemaining)

            # see if should continue
	    if (recTime is None) or (recItem is None):
		# we are done
		break

	    # calc time remaining
            self.curTime = self.curTime + recTime
            timeRemaining = totTime - self.curTime
            
            # see if should continue
	    if (timeRemaining <=0) or (recTime is None) or (recItem is None):
		# we are done
		break


    def Simulate(self,numLists,listLength,contDistract=False,
                 preDistract=50,silent=False,listDef=None):
	"""
	Do a simple simulation of TCM, returning filled recalls and recTimes matrixes.
	"""

        # check listDef
        if not listDef is None:
            numUniqueItems = len(unique1d(listDef))
        else:
            numUniqueItems = listLength

	# allocate for vars to fill
	recalls = zeros([numLists,numUniqueItems])
	recTimes = zeros([numLists,numUniqueItems])
        inputs = zeros([numLists,numUniqueItems*2,numUniqueItems],dtype=float)*nan
        CVs = zeros([numLists,numUniqueItems*2],dtype=float)*nan

        # run one distractor to load in as math dist
        self.Present(updateWeights = False, appendItem = False,
                     isDistract = True, isEncoding = False,
                     customB = 0.0, appendDistract = True)
        
	# train
	# do some distract to clean out
        self.Present(updateWeights = False, appendItem = False, isDistract = True, isEncoding = False, customB = 1.0, appendDistract = False)

	# save the beginning of list item
	self.itemBOL = self.curItem-1
        t_BOL = self.t.copy()
        curItem_BOL = self.curItem

        # do the training
        self.Train(listLength,
                   contDistract = contDistract,
                   listDef = listDef)

        # save the Time Of Test context and weight matrixes
        t_TOT = self.t.copy()
        curItem_TOT = self.curItem
        if self.outputEncoding:
            Mtf_fast_TOT = self.Mtf_fast.copy()
            Mft_fast_TOT = self.Mft_fast.copy()
	
	# test
	if not silent:
	    sys.stderr.write('Lists (%d): ' % (numLists))
	for l in arange(numLists):
	    if not silent and (mod(l+1,50) == 0 or (l+1)==numLists or l==0):
		sys.stderr.write('%d ' % (l+1))
		sys.stderr.flush()

	    # restore recalls 
	    self.recalled = []
	    self.recTime = []
            self.input = []
	    self.CV = []
            self.x = None

            # restore TOT context and weights
            self.t = t_TOT.copy()
            self.curItem = curItem_TOT
            if self.outputEncoding:
                self.Mtf_fast = Mtf_fast_TOT.copy()
                self.Mft_fast = Mft_fast_TOT.copy()
                
	    # test
	    self.Test()

	    # append the recalls
	    recalls[l,:len(self.recalled)] = asarray(self.recalled)+1
	    recTimes[l,:len(self.recTime)] = asarray(self.recTime,'i')
            inputs[l,:len(self.input),:] = asarray(self.input)
            CVs[l,:len(self.CV)] = asarray(self.CV)

	if not silent:
	    sys.stderr.write('\n')
	    sys.stderr.flush()

	# reset to the time of test context
        self.t = t_TOT
        self.curItem = curItem_TOT
	self.x = None

	# return the result
	return (recalls,recTimes,inputs,CVs)


def getValFromDict(dict,key,defval):
    """
    Return the value from a dict if the key exists.
    """
    if dict.has_key(key):
	return dict[key]
    else:
	return defval
